@@ -10,7 +10,7 @@ use crate::{
     errors::HistoryTreeNodeError,
     history_tree_node::*,
     proof_structs::{AppendOnlyProof, MembershipProof, NonMembershipProof},
-    storage::{Storable, V2Storage},
+    storage::{Storable, Storage},
 };
 
 use crate::serialization::to_digest;
@@ -18,7 +18,9 @@ use crate::serialization::to_digest;
 use crate::storage::types::StorageType;
 use crate::{errors::*, history_tree_node::HistoryTreeNode, node_state::*, ARITY, *};
 use async_recursion::async_recursion;
-use std::marker::{PhantomData, Send, Sync};
+use log::{debug, info};
+use std::marker::{Send, Sync};
+use tokio::time::Instant;
 use winter_crypto::Hasher;
 
 use serde::{Deserialize, Serialize};
@@ -29,20 +31,18 @@ use keyed_priority_queue::{Entry, KeyedPriorityQueue};
 pub const DEFAULT_AZKS_KEY: u8 = 1u8;
 /// An append-only zero knowledge set, the data structure used to efficiently implement
 /// a auditable key directory.
-#[derive(Debug, Serialize, Deserialize)]
+#[derive(Debug, Serialize, Deserialize, Eq, PartialEq)]
 #[serde(bound = "")]
-pub struct Azks<H> {
+pub struct Azks {
     /// The location of the root
-    pub root: usize,
+    pub root: u64,
     /// The latest complete epoch
     pub latest_epoch: u64,
     /// The number of nodes ie the size of this tree
-    pub num_nodes: usize, // The size of the tree
-    /// Placeholder
-    pub _h: PhantomData<H>,
+    pub num_nodes: u64, // The size of the tree
 }
 
-impl<H: Hasher> Storable for Azks<H> {
+impl Storable for Azks {
     type Key = u8;
 
     fn data_type() -> StorageType {
@@ -52,30 +52,36 @@ impl<H: Hasher> Storable for Azks<H> {
     fn get_id(&self) -> u8 {
         1u8
     }
+
+    fn get_full_binary_key_id(key: &u8) -> Vec<u8> {
+        vec![StorageType::Azks as u8, *key]
+    }
+
+    fn key_from_full_binary(_bin: &[u8]) -> Result<u8, String> {
+        Ok(1u8)
+    }
 }
 
-unsafe impl<H: Hasher> Sync for Azks<H> {}
+unsafe impl Sync for Azks {}
 
-impl<H: Hasher> Clone for Azks<H> {
+impl Clone for Azks {
     fn clone(&self) -> Self {
         Self {
             root: self.root,
             latest_epoch: self.latest_epoch,
             num_nodes: self.num_nodes,
-            _h: PhantomData,
         }
     }
 }
 
-impl<H: Hasher + Send + Sync> Azks<H> {
+impl Azks {
     /// Creates a new azks
-    pub async fn new<S: V2Storage + Sync + Send>(storage: &S) -> Result<Self, AkdError> {
+    pub async fn new<S: Storage + Sync + Send, H: Hasher>(storage: &S) -> Result<Self, AkdError> {
         let root = get_empty_root::<H, S>(storage, Option::Some(0)).await?;
         let azks = Azks {
             root: 0,
             latest_epoch: 0,
             num_nodes: 1,
-            _h: PhantomData,
         };
 
         root.write_to_storage(storage).await?;
@@ -85,7 +91,7 @@ impl<H: Hasher + Send + Sync> Azks<H> {
 
     /// Inserts a single leaf and is only used for testing, since batching is more efficient.
     /// We just want to make sure batch insertions work correctly and this function is useful for that.
-    pub async fn insert_leaf<S: V2Storage + Sync + Send>(
+    pub async fn insert_leaf<S: Storage + Sync + Send, H: Hasher>(
         &mut self,
         storage: &S,
         label: NodeLabel,
@@ -99,44 +105,110 @@ impl<H: Hasher + Send + Sync> Azks<H> {
 
         let mut root_node = HistoryTreeNode::get_from_storage(storage, NodeKey(self.root)).await?;
         root_node
-            .insert_single_leaf(storage, new_leaf, self.latest_epoch, &mut self.num_nodes)
+            .insert_single_leaf::<_, H>(storage, new_leaf, self.latest_epoch, &mut self.num_nodes)
             .await?;
 
         Ok(())
     }
 
     /// Insert a batch of new leaves
-    pub async fn batch_insert_leaves<S: V2Storage + Sync + Send>(
+    pub async fn batch_insert_leaves<S: Storage + Sync + Send, H: Hasher>(
         &mut self,
         storage: &S,
         insertion_set: Vec<(NodeLabel, H::Digest)>,
     ) -> Result<(), AkdError> {
-        self.batch_insert_leaves_helper(storage, insertion_set, false)
+        self.batch_insert_leaves_helper::<_, H>(storage, insertion_set, false)
             .await
     }
 
+    async fn preload_nodes_for_insertion<S: Storage + Sync + Send, H: Hasher>(
+        &self,
+        storage: &S,
+        insertion_set: &[(NodeLabel, H::Digest)],
+    ) -> Result<u64, AkdError> {
+        let mut load_count: u64 = 0;
+        let mut current_nodes = vec![NodeKey(self.root)];
+
+        let prefixes_set = crate::utils::build_prefixes_set(
+            insertion_set
+                .iter()
+                .map(|(x, _)| *x)
+                .collect::<Vec<NodeLabel>>()
+                .as_ref(),
+        );
+
+        while !current_nodes.is_empty() {
+            let nodes = HistoryTreeNode::batch_get_from_storage(storage, current_nodes).await?;
+            load_count += nodes.len() as u64;
+
+            current_nodes = Vec::<NodeKey>::new();
+            let mut node_states = Vec::<NodeStateKey>::new();
+
+            // This for loop is just getting the keys for states that need to be loaded.
+            for node in &nodes {
+                node_states.push(get_state_map_key(node, node.get_latest_epoch()?));
+            }
+            let states = storage.batch_get::<HistoryNodeState>(node_states).await?;
+            load_count += states.len() as u64;
+
+            // Now that states are loaded in the cache, you can read and access them.
+            // Note, the two for loops are needed because otherwise, you'd be accessing remote storage
+            // individually for each node's state.
+            for node in &nodes {
+                if !prefixes_set.contains(&node.label) {
+                    // Only continue to traverse nodes which are relevant prefixes to insertion_set
+                    continue;
+                }
+
+                for dir in 0..ARITY {
+                    let child = node
+                        .get_child_at_epoch::<S, H>(
+                            storage,
+                            self.latest_epoch,
+                            Direction::Some(dir),
+                        )
+                        .await?;
+
+                    if let Some(child) = child {
+                        current_nodes.push(NodeKey(child.location));
+                    }
+                }
+            }
+        }
+        Ok(load_count)
+    }
+
     /// An azks is built both by the [crate::directory::Directory] and the auditor.
     /// However, both constructions have very minor differences, and the append_only_usage
     /// bool keeps track of this.
-    pub async fn batch_insert_leaves_helper<S: V2Storage + Sync + Send>(
+    pub async fn batch_insert_leaves_helper<S: Storage + Sync + Send, H: Hasher>(
         &mut self,
         storage: &S,
         insertion_set: Vec<(NodeLabel, H::Digest)>,
         append_only_usage: bool,
     ) -> Result<(), AkdError> {
-        self.increment_epoch();
+        let tic = Instant::now();
+        let load_count = self
+            .preload_nodes_for_insertion::<S, H>(storage, &insertion_set)
+            .await?;
+        let toc = Instant::now() - tic;
+        info!(
+            "Preload of tree ({} objects loaded), took {} s",
+            load_count,
+            toc.as_secs_f64()
+        );
 
-        let mut hash_q = KeyedPriorityQueue::<usize, i32>::new();
+        self.increment_epoch();
+        self.preload_nodes_for_insertion::<S, H>(storage, &insertion_set)
+            .await?;
+        let mut hash_q = KeyedPriorityQueue::<u64, i32>::new();
         let mut priorities: i32 = 0;
         let mut root_node = HistoryTreeNode::get_from_storage(storage, NodeKey(self.root)).await?;
         for (label, value) in insertion_set {
             let new_leaf_loc = self.num_nodes;
 
-            let mut new_leaf =
-                get_leaf_node::<H, S>(storage, label, 0, value.as_ref(), 0, self.latest_epoch)
-                    .await?;
-            if append_only_usage {
-                new_leaf = get_leaf_node_without_hashing::<H, S>(
+            let new_leaf = if append_only_usage {
+                get_leaf_node_without_hashing::<H, S>(
                     storage,
                     label,
                     0,
@@ -144,12 +216,17 @@ impl<H: Hasher + Send + Sync> Azks<H> {
                     0,
                     self.latest_epoch,
                 )
-                .await?;
-            }
+                .await?
+            } else {
+                get_leaf_node::<H, S>(storage, label, 0, value.as_ref(), 0, self.latest_epoch)
+                    .await?
+            };
 
+            debug!("BEGIN insert leaf");
             root_node
-                .insert_leaf(storage, new_leaf, self.latest_epoch, &mut self.num_nodes)
+                .insert_leaf::<_, H>(storage, new_leaf, self.latest_epoch, &mut self.num_nodes)
                 .await?;
+            debug!("END insert leaf");
 
             hash_q.push(new_leaf_loc, priorities);
             priorities -= 1;
@@ -160,10 +237,12 @@ impl<H: Hasher + Send + Sync> Azks<H> {
                 .pop()
                 .ok_or(AzksError::PopFromEmptyPriorityQueue(self.latest_epoch))?;
 
-            let mut next_node: HistoryTreeNode<H> =
+            let mut next_node: HistoryTreeNode =
                 HistoryTreeNode::get_from_storage(storage, NodeKey(next_node_loc)).await?;
 
-            next_node.update_hash(storage, self.latest_epoch).await?;
+            next_node
+                .update_hash::<_, H>(storage, self.latest_epoch)
+                .await?;
 
             if !next_node.is_root() {
                 match hash_q.entry(next_node.parent) {
@@ -183,7 +262,7 @@ impl<H: Hasher + Send + Sync> Azks<H> {
 
     /// Returns the Merkle membership proof for the trie as it stood at epoch
     // Assumes the verifier has access to the root at epoch
-    pub async fn get_membership_proof<S: V2Storage + Sync + Send>(
+    pub async fn get_membership_proof<S: Storage + Sync + Send, H: Hasher>(
         &self,
         storage: &S,
         label: NodeLabel,
@@ -198,7 +277,7 @@ impl<H: Hasher + Send + Sync> Azks<H> {
     /// In a compressed trie, the proof consists of the longest prefix
     /// of the label that is included in the trie, as well as its children, to show that
     /// none of the children is equal to the given label.
-    pub async fn get_non_membership_proof<S: V2Storage + Sync + Send>(
+    pub async fn get_non_membership_proof<S: Storage + Sync + Send, H: Hasher>(
         &self,
         storage: &S,
         label: NodeLabel,
@@ -207,20 +286,27 @@ impl<H: Hasher + Send + Sync> Azks<H> {
         let (longest_prefix_membership_proof, lcp_node_id) = self
             .get_membership_proof_and_node(storage, label, epoch)
             .await?;
-        let lcp_node: HistoryTreeNode<H> =
+        let lcp_node: HistoryTreeNode =
             HistoryTreeNode::get_from_storage(storage, NodeKey(lcp_node_id)).await?;
         let longest_prefix = lcp_node.label;
         let mut longest_prefix_children_labels = [NodeLabel::new(0, 0); ARITY];
-        let mut longest_prefix_children_values = [H::hash(&[]); ARITY];
+        let mut longest_prefix_children_values = [crate::utils::empty_node_hash::<H>(); ARITY];
         let state = lcp_node.get_state_at_epoch(storage, epoch).await?;
 
         for (i, child) in state.child_states.iter().enumerate() {
-            let unwrapped_child: HistoryTreeNode<H> =
-                HistoryTreeNode::get_from_storage(storage, NodeKey(child.location)).await?;
-            longest_prefix_children_labels[i] = unwrapped_child.label;
-            longest_prefix_children_values[i] = unwrapped_child
-                .get_value_without_label_at_epoch(storage, epoch)
-                .await?;
+            match child {
+                None => {
+                    continue;
+                }
+                Some(child) => {
+                    let unwrapped_child: HistoryTreeNode =
+                        HistoryTreeNode::get_from_storage(storage, NodeKey(child.location)).await?;
+                    longest_prefix_children_labels[i] = unwrapped_child.label;
+                    longest_prefix_children_values[i] = unwrapped_child
+                        .get_value_without_label_at_epoch::<_, H>(storage, epoch)
+                        .await?;
+                }
+            }
         }
         Ok(NonMembershipProof {
             label,
@@ -240,7 +326,7 @@ impl<H: Hasher + Send + Sync> Azks<H> {
     /// **RESTRICTIONS**: Note that `start_epoch` and `end_epoch` are valid only when the following are true
     /// * `start_epoch` <= `end_epoch`
     /// * `start_epoch` and `end_epoch` are both existing epochs of this AZKS
-    pub async fn get_append_only_proof<S: V2Storage + Sync + Send>(
+    pub async fn get_append_only_proof<S: Storage + Sync + Send, H: Hasher>(
         &self,
         storage: &S,
         start_epoch: u64,
@@ -251,7 +337,7 @@ impl<H: Hasher + Send + Sync> Azks<H> {
         // between these epochs.
         let node = HistoryTreeNode::get_from_storage(storage, NodeKey(self.root)).await?;
         let (unchanged, leaves) = self
-            .get_append_only_proof_helper(storage, node, start_epoch, end_epoch)
+            .get_append_only_proof_helper::<_, H>(storage, node, start_epoch, end_epoch)
             .await?;
         Ok(AppendOnlyProof {
             inserted: leaves,
@@ -260,10 +346,10 @@ impl<H: Hasher + Send + Sync> Azks<H> {
     }
 
     #[async_recursion]
-    async fn get_append_only_proof_helper<S: V2Storage + Sync + Send>(
+    async fn get_append_only_proof_helper<S: Storage + Sync + Send, H: Hasher>(
         &self,
         storage: &S,
-        node: HistoryTreeNode<H>,
+        node: HistoryTreeNode,
         start_epoch: u64,
         end_epoch: u64,
     ) -> Result<AppendOnlyHelper<H::Digest>, AkdError> {
@@ -277,7 +363,7 @@ impl<H: Hasher + Send + Sync> Azks<H> {
 
             unchanged.push((
                 node.label,
-                node.get_value_without_label_at_epoch(storage, node.get_latest_epoch()?)
+                node.get_value_without_label_at_epoch::<_, H>(storage, node.get_latest_epoch()?)
                     .await?,
             ));
             return Ok((unchanged, leaves));
@@ -289,7 +375,7 @@ impl<H: Hasher + Send + Sync> Azks<H> {
         if node.is_leaf() {
             leaves.push((
                 node.label,
-                node.get_value_without_label_at_epoch(storage, node.get_latest_epoch()?)
+                node.get_value_without_label_at_epoch::<_, H>(storage, node.get_latest_epoch()?)
                     .await?,
             ));
         } else {
@@ -300,19 +386,27 @@ impl<H: Hasher + Send + Sync> Azks<H> {
                 .iter()
                 .map(|x| x.clone())
             {
-                if child_node_state.dummy_marker == DummyChildState::Dummy {
-                    continue;
-                } else {
-                    let child_node = HistoryTreeNode::get_from_storage(
-                        storage,
-                        NodeKey(child_node_state.location),
-                    )
-                    .await?;
-                    let mut rec_output = self
-                        .get_append_only_proof_helper(storage, child_node, start_epoch, end_epoch)
+                match child_node_state {
+                    None => {
+                        continue;
+                    }
+                    Some(child_node_state) => {
+                        let child_node = HistoryTreeNode::get_from_storage(
+                            storage,
+                            NodeKey(child_node_state.location),
+                        )
                         .await?;
-                    unchanged.append(&mut rec_output.0);
-                    leaves.append(&mut rec_output.1);
+                        let mut rec_output = self
+                            .get_append_only_proof_helper::<_, H>(
+                                storage,
+                                child_node,
+                                start_epoch,
+                                end_epoch,
+                            )
+                            .await?;
+                        unchanged.append(&mut rec_output.0);
+                        leaves.append(&mut rec_output.1);
+                    }
                 }
             }
         }
@@ -321,25 +415,25 @@ impl<H: Hasher + Send + Sync> Azks<H> {
 
     // FIXME: these functions below should be moved into higher-level API
     /// Gets the root hash for this azks
-    pub async fn get_root_hash<S: V2Storage + Sync + Send>(
+    pub async fn get_root_hash<S: Storage + Sync + Send, H: Hasher>(
         &self,
         storage: &S,
     ) -> Result<H::Digest, HistoryTreeNodeError> {
-        self.get_root_hash_at_epoch(storage, self.get_latest_epoch())
+        self.get_root_hash_at_epoch::<_, H>(storage, self.get_latest_epoch())
             .await
     }
 
     /// Gets the root hash of the tree at a epoch.
     /// Since this is accessing the root node and the root node exists at all epochs that
     /// the azks does, this would never be called at an epoch before the birth of the root node.
-    pub async fn get_root_hash_at_epoch<S: V2Storage + Sync + Send>(
+    pub async fn get_root_hash_at_epoch<S: Storage + Sync + Send, H: Hasher>(
         &self,
         storage: &S,
         epoch: u64,
     ) -> Result<H::Digest, HistoryTreeNodeError> {
-        let root_node: HistoryTreeNode<H> =
+        let root_node: HistoryTreeNode =
             HistoryTreeNode::get_from_storage(storage, NodeKey(self.root)).await?;
-        root_node.get_value_at_epoch(storage, epoch).await
+        root_node.get_value_at_epoch::<_, H>(storage, epoch).await
     }
 
     /// Gets the latest epoch of this azks. If an update aka epoch transition
@@ -356,17 +450,17 @@ impl<H: Hasher + Send + Sync> Azks<H> {
     /// This function returns the node location for the node whose label is the longest common
     /// prefix for the queried label. It also returns a membership proof for said label.
     /// This is meant to be used in both, getting membership proofs and getting non-membership proofs.
-    pub async fn get_membership_proof_and_node<S: V2Storage + Sync + Send>(
+    pub async fn get_membership_proof_and_node<S: Storage + Sync + Send, H: Hasher>(
         &self,
         storage: &S,
         label: NodeLabel,
         epoch: u64,
-    ) -> Result<(MembershipProof<H>, usize), AkdError> {
+    ) -> Result<(MembershipProof<H>, u64), AkdError> {
         let mut parent_labels = Vec::<NodeLabel>::new();
         let mut sibling_labels = Vec::<[NodeLabel; ARITY - 1]>::new();
         let mut sibling_hashes = Vec::<[H::Digest; ARITY - 1]>::new();
         let mut dirs = Vec::<Direction>::new();
-        let mut curr_node: HistoryTreeNode<H> =
+        let mut curr_node: HistoryTreeNode =
             HistoryTreeNode::get_from_storage(storage, NodeKey(self.root)).await?;
         let mut dir = curr_node.label.get_dir(label);
         let mut equal = label == curr_node.label;
@@ -381,23 +475,27 @@ impl<H: Hasher + Send + Sync> Azks<H> {
             let mut count = 0;
             let direction = dir.ok_or(AkdError::NoDirectionError)?;
             let next_state = curr_state.get_child_state_in_dir(direction);
-            if next_state.dummy_marker == DummyChildState::Dummy {
+            if next_state == None {
                 break;
             }
             for i in 0..ARITY {
                 if i != dir.ok_or(AkdError::NoDirectionError)? {
-                    labels[count] = curr_state.child_states[i].label;
-                    hashes[count] = to_digest::<H>(&curr_state.child_states[i].hash_val).unwrap();
+                    labels[count] =
+                        optional_history_child_state_to_label(&curr_state.child_states[i]);
+                    hashes[count] = to_digest::<H>(&optional_history_child_state_to_hash::<H>(
+                        &curr_state.child_states[i],
+                    ))
+                    .unwrap();
                     count += 1;
                 }
             }
             sibling_labels.push(labels);
             sibling_hashes.push(hashes);
-            let new_curr_node: HistoryTreeNode<H> = HistoryTreeNode::get_from_storage(
+            let new_curr_node: HistoryTreeNode = HistoryTreeNode::get_from_storage(
                 storage,
                 NodeKey(
                     curr_node
-                        .get_child_location_at_epoch(storage, epoch, dir)
+                        .get_child_location_at_epoch::<_, H>(storage, epoch, dir)
                         .await?,
                 ),
             )
@@ -407,7 +505,7 @@ impl<H: Hasher + Send + Sync> Azks<H> {
             equal = label == curr_node.label;
         }
         if !equal {
-            let new_curr_node: HistoryTreeNode<H> =
+            let new_curr_node: HistoryTreeNode =
                 HistoryTreeNode::get_from_storage(storage, NodeKey(prev_node)).await?;
             curr_node = new_curr_node;
 
@@ -418,7 +516,7 @@ impl<H: Hasher + Send + Sync> Azks<H> {
         }
 
         let hash_val = curr_node
-            .get_value_without_label_at_epoch(storage, epoch)
+            .get_value_without_label_at_epoch::<_, H>(storage, epoch)
             .await?;
 
         Ok((
@@ -452,12 +550,12 @@ mod tests {
     type Blake3 = Blake3_256<BaseElement>;
     type Blake3Digest = <Blake3_256<winter_math::fields::f128::BaseElement> as Hasher>::Digest;
 
-    #[actix_rt::test]
+    #[tokio::test]
     async fn test_batch_insert_basic() -> Result<(), AkdError> {
         let mut rng = OsRng;
         let num_nodes = 10;
-        let db = storage::V2FromV1StorageWrapper::new(AsyncInMemoryDatabase::new());
-        let mut azks1 = Azks::<Blake3>::new(&db).await?;
+        let db = AsyncInMemoryDatabase::new();
+        let mut azks1 = Azks::new::<_, Blake3>(&db).await?;
 
         let mut insertion_set: Vec<(NodeLabel, Blake3Digest)> = vec![];
 
@@ -467,29 +565,31 @@ mod tests {
             rng.fill_bytes(&mut input);
             let val = Blake3::hash(&input);
             insertion_set.push((node, val));
-            azks1.insert_leaf(&db, node, val).await?;
+            azks1.insert_leaf::<_, Blake3>(&db, node, val).await?;
         }
 
-        let db2 = storage::V2FromV1StorageWrapper::new(AsyncInMemoryDatabase::new());
-        let mut azks2 = Azks::<Blake3>::new(&db2).await?;
+        let db2 = AsyncInMemoryDatabase::new();
+        let mut azks2 = Azks::new::<_, Blake3>(&db2).await?;
 
-        azks2.batch_insert_leaves(&db2, insertion_set).await?;
+        azks2
+            .batch_insert_leaves::<_, Blake3>(&db2, insertion_set)
+            .await?;
 
         assert_eq!(
-            azks1.get_root_hash(&db).await?,
-            azks2.get_root_hash(&db2).await?,
+            azks1.get_root_hash::<_, Blake3>(&db).await?,
+            azks2.get_root_hash::<_, Blake3>(&db2).await?,
             "Batch insert doesn't match individual insert"
         );
 
         Ok(())
     }
 
-    #[actix_rt::test]
+    #[tokio::test]
     async fn test_insert_permuted() -> Result<(), AkdError> {
         let num_nodes = 10;
         let mut rng = OsRng;
-        let db = storage::V2FromV1StorageWrapper::new(AsyncInMemoryDatabase::new());
-        let mut azks1 = Azks::<Blake3>::new(&db).await?;
+        let db = AsyncInMemoryDatabase::new();
+        let mut azks1 = Azks::new::<_, Blake3>(&db).await?;
         let mut insertion_set: Vec<(NodeLabel, Blake3Digest)> = vec![];
 
         for _ in 0..num_nodes {
@@ -498,27 +598,29 @@ mod tests {
             rng.fill_bytes(&mut input);
             let input = Blake3Digest::new(input);
             insertion_set.push((node, input));
-            azks1.insert_leaf(&db, node, input).await?;
+            azks1.insert_leaf::<_, Blake3>(&db, node, input).await?;
         }
 
         // Try randomly permuting
         insertion_set.shuffle(&mut rng);
 
-        let db2 = storage::V2FromV1StorageWrapper::new(AsyncInMemoryDatabase::new());
-        let mut azks2 = Azks::<Blake3>::new(&db2).await?;
+        let db2 = AsyncInMemoryDatabase::new();
+        let mut azks2 = Azks::new::<_, Blake3>(&db2).await?;
 
-        azks2.batch_insert_leaves(&db2, insertion_set).await?;
+        azks2
+            .batch_insert_leaves::<_, Blake3>(&db2, insertion_set)
+            .await?;
 
         assert_eq!(
-            azks1.get_root_hash(&db).await?,
-            azks2.get_root_hash(&db2).await?,
+            azks1.get_root_hash::<_, Blake3>(&db).await?,
+            azks2.get_root_hash::<_, Blake3>(&db2).await?,
             "Batch insert doesn't match individual insert"
         );
 
         Ok(())
     }
 
-    #[actix_rt::test]
+    #[tokio::test]
     async fn test_membership_proof_permuted() -> Result<(), AkdError> {
         let num_nodes = 10;
         let mut rng = OsRng;
@@ -535,20 +637,21 @@ mod tests {
 
         // Try randomly permuting
         insertion_set.shuffle(&mut rng);
-        let db = storage::V2FromV1StorageWrapper::new(AsyncInMemoryDatabase::new());
-        let mut azks = Azks::<Blake3>::new(&db).await?;
-        azks.batch_insert_leaves(&db, insertion_set.clone()).await?;
+        let db = AsyncInMemoryDatabase::new();
+        let mut azks = Azks::new::<_, Blake3>(&db).await?;
+        azks.batch_insert_leaves::<_, Blake3>(&db, insertion_set.clone())
+            .await?;
 
         let proof = azks
             .get_membership_proof(&db, insertion_set[0].0, 1)
             .await?;
 
-        verify_membership::<Blake3>(azks.get_root_hash(&db).await?, &proof)?;
+        verify_membership::<Blake3>(azks.get_root_hash::<_, Blake3>(&db).await?, &proof)?;
 
         Ok(())
     }
 
-    #[actix_rt::test]
+    #[tokio::test]
     async fn test_membership_proof_failing() -> Result<(), AkdError> {
         let num_nodes = 10;
         let mut rng = OsRng;
@@ -565,9 +668,10 @@ mod tests {
 
         // Try randomly permuting
         insertion_set.shuffle(&mut rng);
-        let db = storage::V2FromV1StorageWrapper::new(AsyncInMemoryDatabase::new());
-        let mut azks = Azks::<Blake3>::new(&db).await?;
-        azks.batch_insert_leaves(&db, insertion_set.clone()).await?;
+        let db = AsyncInMemoryDatabase::new();
+        let mut azks = Azks::new::<_, Blake3>(&db).await?;
+        azks.batch_insert_leaves::<_, Blake3>(&db, insertion_set.clone())
+            .await?;
 
         let mut proof = azks
             .get_membership_proof(&db, insertion_set[0].0, 1)
@@ -582,34 +686,36 @@ mod tests {
             dirs: proof.dirs,
         };
         assert!(
-            !verify_membership::<Blake3>(azks.get_root_hash(&db).await?, &proof).is_ok(),
+            !verify_membership::<Blake3>(azks.get_root_hash::<_, Blake3>(&db).await?, &proof)
+                .is_ok(),
             "Membership proof does verifies, despite being wrong"
         );
 
         Ok(())
     }
 
-    #[actix_rt::test]
+    #[tokio::test]
     async fn test_membership_proof_intermediate() -> Result<(), AkdError> {
-        let db = storage::V2FromV1StorageWrapper::new(AsyncInMemoryDatabase::new());
+        let db = AsyncInMemoryDatabase::new();
         let mut insertion_set: Vec<(NodeLabel, Blake3Digest)> = vec![];
         insertion_set.push((NodeLabel::new(0b0, 64), Blake3::hash(&[])));
         insertion_set.push((NodeLabel::new(0b1 << 63, 64), Blake3::hash(&[])));
         insertion_set.push((NodeLabel::new(0b11 << 62, 64), Blake3::hash(&[])));
         insertion_set.push((NodeLabel::new(0b01 << 62, 64), Blake3::hash(&[])));
         insertion_set.push((NodeLabel::new(0b111 << 61, 64), Blake3::hash(&[])));
-        let mut azks = Azks::<Blake3>::new(&db).await?;
-        azks.batch_insert_leaves(&db, insertion_set).await?;
+        let mut azks = Azks::new::<_, Blake3>(&db).await?;
+        azks.batch_insert_leaves::<_, Blake3>(&db, insertion_set)
+            .await?;
         let search_label = NodeLabel::new(0b1111 << 60, 64);
         let proof = azks.get_non_membership_proof(&db, search_label, 1).await?;
         assert!(
-            verify_nonmembership::<Blake3>(azks.get_root_hash(&db).await?, &proof)?,
+            verify_nonmembership::<Blake3>(azks.get_root_hash::<_, Blake3>(&db).await?, &proof)?,
             "Nonmembership proof does not verify"
         );
         Ok(())
     }
 
-    #[actix_rt::test]
+    #[tokio::test]
     async fn test_nonmembership_proof() -> Result<(), AkdError> {
         let num_nodes = 10;
         let mut rng = OsRng;
@@ -623,36 +729,41 @@ mod tests {
             let input = Blake3Digest::new(input);
             insertion_set.push((node, input));
         }
-        let db = storage::V2FromV1StorageWrapper::new(AsyncInMemoryDatabase::new());
-        let mut azks = Azks::<Blake3>::new(&db).await?;
+        let db = AsyncInMemoryDatabase::new();
+        let mut azks = Azks::new::<_, Blake3>(&db).await?;
         let search_label = insertion_set[num_nodes - 1].0;
-        azks.batch_insert_leaves(&db, insertion_set.clone()[0..num_nodes - 1].to_vec())
-            .await?;
+        azks.batch_insert_leaves::<_, Blake3>(
+            &db,
+            insertion_set.clone()[0..num_nodes - 1].to_vec(),
+        )
+        .await?;
         let proof = azks.get_non_membership_proof(&db, search_label, 1).await?;
 
         assert!(
-            verify_nonmembership::<Blake3>(azks.get_root_hash(&db).await?, &proof)?,
+            verify_nonmembership::<Blake3>(azks.get_root_hash::<_, Blake3>(&db).await?, &proof)?,
             "Nonmembership proof does not verify"
         );
 
         Ok(())
     }
 
-    #[actix_rt::test]
+    #[tokio::test]
     async fn test_append_only_proof_very_tiny() -> Result<(), AkdError> {
-        let db = storage::V2FromV1StorageWrapper::new(AsyncInMemoryDatabase::new());
-        let mut azks = Azks::<Blake3>::new(&db).await?;
+        let db = AsyncInMemoryDatabase::new();
+        let mut azks = Azks::new::<_, Blake3>(&db).await?;
 
         let mut insertion_set_1: Vec<(NodeLabel, Blake3Digest)> = vec![];
         insertion_set_1.push((NodeLabel::new(0b0, 64), Blake3::hash(&[])));
-        azks.batch_insert_leaves(&db, insertion_set_1).await?;
-        let start_hash = azks.get_root_hash(&db).await?;
+        azks.batch_insert_leaves::<_, Blake3>(&db, insertion_set_1)
+            .await?;
+        let start_hash = azks.get_root_hash::<_, Blake3>(&db).await?;
 
         let mut insertion_set_2: Vec<(NodeLabel, Blake3Digest)> = vec![];
         insertion_set_2.push((NodeLabel::new(0b01 << 62, 64), Blake3::hash(&[])));
 
-        azks.batch_insert_leaves(&db, insertion_set_2).await?;
-        let end_hash = azks.get_root_hash(&db).await?;
+        azks.batch_insert_leaves::<_, Blake3>(&db, insertion_set_2)
+            .await?;
+        let end_hash = azks.get_root_hash::<_, Blake3>(&db).await?;
 
         let proof = azks.get_append_only_proof(&db, 1, 2).await?;
 
@@ -660,23 +771,25 @@ mod tests {
         Ok(())
     }
 
-    #[actix_rt::test]
+    #[tokio::test]
     async fn test_append_only_proof_tiny() -> Result<(), AkdError> {
-        let db = storage::V2FromV1StorageWrapper::new(AsyncInMemoryDatabase::new());
-        let mut azks = Azks::<Blake3>::new(&db).await?;
+        let db = AsyncInMemoryDatabase::new();
+        let mut azks = Azks::new::<_, Blake3>(&db).await?;
 
         let mut insertion_set_1: Vec<(NodeLabel, Blake3Digest)> = vec![];
         insertion_set_1.push((NodeLabel::new(0b0, 64), Blake3::hash(&[])));
         insertion_set_1.push((NodeLabel::new(0b1 << 63, 64), Blake3::hash(&[])));
-        azks.batch_insert_leaves(&db, insertion_set_1).await?;
-        let start_hash = azks.get_root_hash(&db).await?;
+        azks.batch_insert_leaves::<_, Blake3>(&db, insertion_set_1)
+            .await?;
+        let start_hash = azks.get_root_hash::<_, Blake3>(&db).await?;
 
         let mut insertion_set_2: Vec<(NodeLabel, Blake3Digest)> = vec![];
         insertion_set_2.push((NodeLabel::new(0b01 << 62, 64), Blake3::hash(&[])));
         insertion_set_2.push((NodeLabel::new(0b111 << 61, 64), Blake3::hash(&[])));
 
-        azks.batch_insert_leaves(&db, insertion_set_2).await?;
-        let end_hash = azks.get_root_hash(&db).await?;
+        azks.batch_insert_leaves::<_, Blake3>(&db, insertion_set_2)
+            .await?;
+        let end_hash = azks.get_root_hash::<_, Blake3>(&db).await?;
 
         let proof = azks.get_append_only_proof(&db, 1, 2).await?;
 
@@ -684,7 +797,7 @@ mod tests {
         Ok(())
     }
 
-    #[actix_rt::test]
+    #[tokio::test]
     async fn test_append_only_proof() -> Result<(), AkdError> {
         let num_nodes = 10;
         let mut rng = OsRng;
@@ -699,12 +812,12 @@ mod tests {
             insertion_set_1.push((node, input));
         }
 
-        let db = storage::V2FromV1StorageWrapper::new(AsyncInMemoryDatabase::new());
-        let mut azks = Azks::<Blake3>::new(&db).await?;
-        azks.batch_insert_leaves(&db, insertion_set_1.clone())
+        let db = AsyncInMemoryDatabase::new();
+        let mut azks = Azks::new::<_, Blake3>(&db).await?;
+        azks.batch_insert_leaves::<_, Blake3>(&db, insertion_set_1.clone())
             .await?;
 
-        let start_hash = azks.get_root_hash(&db).await?;
+        let start_hash = azks.get_root_hash::<_, Blake3>(&db).await?;
 
         let mut insertion_set_2: Vec<(NodeLabel, Blake3Digest)> = vec![];
 
@@ -716,7 +829,7 @@ mod tests {
             insertion_set_2.push((node, input));
         }
 
-        azks.batch_insert_leaves(&db, insertion_set_2.clone())
+        azks.batch_insert_leaves::<_, Blake3>(&db, insertion_set_2.clone())
             .await?;
 
         let mut insertion_set_3: Vec<(NodeLabel, Blake3Digest)> = vec![];
@@ -729,10 +842,10 @@ mod tests {
             insertion_set_3.push((node, input));
         }
 
-        azks.batch_insert_leaves(&db, insertion_set_3.clone())
+        azks.batch_insert_leaves::<_, Blake3>(&db, insertion_set_3.clone())
             .await?;
 
-        let end_hash = azks.get_root_hash(&db).await?;
+        let end_hash = azks.get_root_hash::<_, Blake3>(&db).await?;
 
         let proof = azks.get_append_only_proof(&db, 1, 3).await?;
 
